export function concat(iterable) {
    const chunks = Array.isArray(iterable) ? iterable : Array.from(iterable);
    let size = 0;
    for (let i = 0; i < chunks.length; i++) {
        size += chunks[i].length;
    }
    const concated = new Uint8Array(size);
    for (let i = 0, offset = 0; i < chunks.length; i++) {
        concated.set(chunks[i], offset);
        offset += chunks[i].length;
    }
    return concated;
}
export async function asyncConcat(iterable, maxSize = Infinity) {
    let chunks = [];
    let size = 0;
    let item;
    while (!(item = await iterable.next()).done) {
        if ((size += item.value.length) > maxSize) {
            throw new Error(`Size exceeded ${maxSize} bytes`);
        }
        chunks.push(item.value);
    }
    return concat(chunks);
}
export class Splitable {
    constructor(rawIterable) {
        this.remain = undefined;
        this.done = false;
        this.rawIterator = rawIterable[Symbol.asyncIterator]();
    }
    async hasValue() {
        return this.next().then((item) => {
            if (item.done) {
                return false;
            }
            this.remain = item.value;
            return true;
        });
    }
    async next() {
        var _a;
        if (this.done) {
            return { done: true, value: undefined };
        }
        if ((_a = this.remain) === null || _a === void 0 ? void 0 : _a.length) {
            const item = { done: false, value: this.remain };
            this.remain = undefined;
            return item;
        }
        return this.rawIterator.next().then((item) => {
            if (item.done) {
                this.done = true;
                this.remain = undefined;
                return item;
            }
            if (item.value.length) {
                return item;
            }
            return this.next();
        });
    }
    async return(value) {
        if (this.rawIterator.return) {
            return this.rawIterator.return(value).then((item) => {
                if (item.done) {
                    this.done = true;
                    this.remain = undefined;
                }
                return item;
            });
        }
        this.done = true;
        this.remain = undefined;
        return { done: true, value };
    }
    async throw(error) {
        if (this.rawIterator.throw) {
            return this.rawIterator.throw(error).then((item) => {
                if (item.done) {
                    this.done = true;
                    this.remain = undefined;
                }
                return item;
            });
        }
        this.done = true;
        this.remain = undefined;
        return Promise.reject(error);
    }
    splitLine() {
        const splitable = this;
        let done = false;
        return {
            async next() {
                if (done) {
                    return { done: true, value: undefined };
                }
                return splitable.next().then((item) => {
                    if (item.done) {
                        done = true;
                        return item;
                    }
                    const lf = item.value.indexOf(10);
                    if (lf !== -1) {
                        done = true;
                        splitable.remain = item.value.subarray(lf + 1);
                        const cr = Math.max(0, lf - 1);
                        return {
                            done: false,
                            value: item.value.subarray(0, item.value[cr] === 13 ? cr : lf),
                        };
                    }
                    if (item.value[item.value.length - 1] !== 13) {
                        return item;
                    }
                    return splitable.next().then((nextItem) => {
                        if (nextItem.done) {
                            done = true;
                            return item;
                        }
                        else if (nextItem.value[0] === 10) {
                            done = true;
                            splitable.remain = nextItem.value.subarray(1);
                            return {
                                done: false,
                                value: item.value.subarray(0, item.value.length - 1),
                            };
                        }
                        else {
                            splitable.remain = nextItem.value;
                            return item;
                        }
                    });
                });
            },
            async return(value) {
                return splitable.return(value).then((item) => {
                    if (item.done) {
                        done = true;
                    }
                    return item;
                });
            },
            async throw(error) {
                return splitable.throw(error).then((item) => {
                    if (item.done) {
                        done = true;
                    }
                    return item;
                });
            },
            [Symbol.asyncIterator]() {
                return this;
            },
        };
    }
    splitSize(size) {
        if (typeof size === "number" && size >= 0 && Math.floor(size) === size) {
            const splitable = this;
            let done = false;
            return {
                async next() {
                    if (done) {
                        return { done: true, value: undefined };
                    }
                    return splitable.next().then((item) => {
                        if (item.done) {
                            done = true;
                            return item;
                        }
                        if (size > item.value.length) {
                            size -= item.value.length;
                            return item;
                        }
                        done = true;
                        splitable.remain = item.value.subarray(size);
                        return { done: false, value: item.value.subarray(0, size) };
                    });
                },
                async return(value) {
                    return splitable.return(value).then((item) => {
                        if (item.done) {
                            done = true;
                        }
                        return item;
                    });
                },
                async throw(error) {
                    return splitable.throw(error).then((item) => {
                        if (item.done) {
                            done = true;
                        }
                        return item;
                    });
                },
                [Symbol.asyncIterator]() {
                    return this;
                },
            };
        }
        throw new Error("Invalid size");
    }
    splitBeforeNeedle(needle) {
        if (needle instanceof Uint8Array && needle.length) {
            const splitable = this;
            let done = false;
            const move = new Array(256).fill(needle.length);
            for (let i = 0; i < needle.length - 1; i++) {
                move[needle[i]] = needle.length - 1 - i;
            }
            return {
                async next() {
                    if (done) {
                        return { done: true, value: undefined };
                    }
                    const haystack = await splitable.eagerLoad(needle.length);
                    if (haystack.length < needle.length) {
                        done = true;
                        return haystack.length === 0 ? { done: true, value: undefined } : { done: false, value: haystack };
                    }
                    let index = -1;
                    let offset = 0;
                    while (offset <= haystack.length - needle.length) {
                        let i = offset;
                        let j = 0;
                        while (j < needle.length && haystack[i] === needle[j]) {
                            i++;
                            j++;
                        }
                        if (j === needle.length) {
                            index = offset;
                            break;
                        }
                        offset += move[haystack[offset + needle.length - 1]];
                    }
                    if (index !== -1) {
                        done = true;
                        splitable.remain = haystack.subarray(index + needle.length);
                        return { done: false, value: haystack.subarray(0, index) };
                    }
                    splitable.remain = haystack.subarray(offset);
                    return { done: false, value: haystack.subarray(0, offset) };
                },
                async return(value) {
                    return splitable.return(value).then((item) => {
                        if (item.done) {
                            done = true;
                        }
                        return item;
                    });
                },
                async throw(error) {
                    return splitable.throw(error).then((item) => {
                        if (item.done) {
                            done = true;
                        }
                        return item;
                    });
                },
                [Symbol.asyncIterator]() {
                    return this;
                },
            };
        }
        throw new Error("Invalid needle");
    }
    splitChunkedHTTPBody() {
        const splitable = this;
        let done = false;
        let chunkSize = 0;
        let chunkRealSize = 0;
        let chunkIterator;
        const ascii2decimal = new Map([
            [48, 0],
            [49, 1],
            [50, 2],
            [51, 3],
            [52, 4],
            [53, 5],
            [54, 6],
            [55, 7],
            [56, 8],
            [57, 9],
            [65, 10],
            [66, 11],
            [67, 12],
            [68, 13],
            [69, 14],
            [70, 15],
            [97, 10],
            [98, 11],
            [99, 12],
            [100, 13],
            [101, 14],
            [102, 15],
        ]);
        return {
            async next() {
                if (done) {
                    return { done: true, value: undefined };
                }
                if (chunkSize === 0) {
                    const chunkLine = await splitable.readLine(16384);
                    const hexs = [];
                    for (const byte of chunkLine) {
                        if (ascii2decimal.has(byte)) {
                            hexs.unshift(ascii2decimal.get(byte));
                        }
                        else {
                            break;
                        }
                    }
                    if (hexs.length === 0) {
                        throw new Error("Missing chunk size");
                    }
                    if (hexs.length > 14 || (hexs.length === 14 && hexs[13] > 1)) {
                        throw new Error("Chunk size exceeded Number.MAX_SAFE_INTEGER");
                    }
                    chunkSize = 0;
                    for (let i = 0; i < hexs.length; i++) {
                        chunkSize += hexs[i] * Math.pow(16, i);
                    }
                    if (chunkSize === 0) {
                        done = true;
                        return splitable.readLine(0).then(() => ({ done: true, value: undefined }));
                    }
                    chunkIterator = splitable.splitSize(chunkSize);
                }
                const item = await chunkIterator.next();
                if (!item.done) {
                    chunkRealSize += item.value.length;
                    return item;
                }
                if (chunkRealSize !== chunkSize) {
                    return Promise.reject(new Error("Don't have enough size of data"));
                }
                chunkSize = 0;
                chunkRealSize = 0;
                return splitable.readLine(0).then(() => this.next());
            },
            async return(value) {
                return splitable.return(value).then((item) => {
                    if (item.done) {
                        done = true;
                    }
                    return item;
                });
            },
            async throw(error) {
                return splitable.throw(error).then((item) => {
                    if (item.done) {
                        done = true;
                    }
                    return item;
                });
            },
            [Symbol.asyncIterator]() {
                return this;
            },
        };
    }
    splitChunkedStream(littleEndian = false) {
        const splitable = this;
        let done = false;
        let chunkSize = 0;
        let chunkRealSize = 0;
        let chunkIterator;
        return {
            async next() {
                if (done) {
                    return { done: true, value: undefined };
                }
                if (chunkSize === 0) {
                    chunkSize = new DataView((await splitable.readEnoughSize(4)).buffer).getUint32(0, littleEndian);
                    if (chunkSize === 0) {
                        done = true;
                        return { done: true, value: undefined };
                    }
                    chunkIterator = splitable.splitSize(chunkSize);
                }
                const item = await chunkIterator.next();
                if (!item.done) {
                    chunkRealSize += item.value.length;
                    return item;
                }
                if (chunkRealSize !== chunkSize) {
                    throw new Error("Don't have enough size of data");
                }
                chunkSize = 0;
                chunkRealSize = 0;
                return this.next();
            },
            async return(value) {
                return splitable.return(value).then((item) => {
                    if (item.done) {
                        done = true;
                    }
                    return item;
                });
            },
            async throw(error) {
                return splitable.throw(error).then((item) => {
                    if (item.done) {
                        done = true;
                    }
                    return item;
                });
            },
            [Symbol.asyncIterator]() {
                return this;
            },
        };
    }
    async readLine(maxSize = Infinity) {
        return asyncConcat(this.splitLine(), maxSize);
    }
    async readSize(size) {
        return asyncConcat(this.splitSize(size));
    }
    async readEnoughSize(size) {
        const concated = await asyncConcat(this.splitSize(size));
        if (concated.length < size) {
            throw new Error("Don't have enough size of data");
        }
        return concated;
    }
    async readBeforeNeedle(needle, maxSize = Infinity) {
        return asyncConcat(this.splitBeforeNeedle(needle), maxSize);
    }
    async readChunkedHTTPBody(maxSize = Infinity) {
        return asyncConcat(this.splitChunkedHTTPBody(), maxSize);
    }
    async readChunkedStream(littleEndian = false, maxSize = Infinity) {
        return asyncConcat(this.splitChunkedStream(littleEndian), maxSize);
    }
    async eagerLoad(expect) {
        const item = await this.next();
        if (item.done) {
            return new Uint8Array(0);
        }
        if (item.value.length >= expect) {
            return item.value;
        }
        let chunks = [item.value];
        let chunksLen = item.value.length;
        while (chunksLen < expect) {
            const moreItem = await this.next();
            if (moreItem.done) {
                break;
            }
            chunks.push(moreItem.value);
            chunksLen += moreItem.value.length;
        }
        const concated = new Uint8Array(chunksLen);
        for (let i = 0, offset = 0; i < chunks.length; i++) {
            concated.set(chunks[i], offset);
            offset += chunks[i].length;
        }
        return concated;
    }
    [Symbol.asyncIterator]() {
        return this;
    }
}
